@@ -1,10 +1,25 @@
-__version__ = "0.1.11"
-from qibo import callbacks, gates, hamiltonians, models, optimizers, parallel, solvers
+import importlib.metadata as im
+
+__version__ = im.version(__package__)
+
+from qibo import (
+    callbacks,
+    gates,
+    hamiltonians,
+    models,
+    optimizers,
+    parallel,
+    parameter,
+    result,
+    solvers,
+)
 from qibo.backends import (
+    construct_backend,
     get_backend,
     get_device,
     get_precision,
     get_threads,
+    list_available_backends,
     matrices,
     set_backend,
     set_device,
@@ -17,3 +32,4 @@ from qibo.config import (
     set_batch_size,
     set_metropolis_threshold,
 )
+from qibo.models.circuit import Circuit
