@@ -1,13 +1,16 @@
 import collections
+import math
 
 import numpy as np
+from scipy import sparse
+from scipy.linalg import block_diag
 
 from qibo import __version__
 from qibo.backends import einsum_utils
 from qibo.backends.abstract import Backend
 from qibo.backends.npmatrices import NumpyMatrices
 from qibo.config import log, raise_error
-from qibo.states import CircuitResult
+from qibo.result import CircuitResult, MeasurementOutcomes, QuantumState
 
 
 class NumpyBackend(Backend):
@@ -85,9 +88,15 @@ class NumpyBackend(Backend):
         state[0, 0] = 1
         return state
 
+    def identity_density_matrix(self, nqubits, normalize: bool = True):
+        state = self.np.eye(2**nqubits, dtype=self.dtype)
+        if normalize is True:
+            state /= 2**nqubits
+        return state
+
     def plus_state(self, nqubits):
         state = self.np.ones(2**nqubits, dtype=self.dtype)
-        state /= self.np.sqrt(2**nqubits)
+        state /= math.sqrt(2**nqubits)
         return state
 
     def plus_density_matrix(self, nqubits):
@@ -95,26 +104,39 @@ class NumpyBackend(Backend):
         state /= 2**nqubits
         return state
 
-    def asmatrix(self, gate):
+    def matrix(self, gate):
         """Convert a gate to its matrix representation in the computational basis."""
         name = gate.__class__.__name__
-        return getattr(self.matrices, name)
+        _matrix = getattr(self.matrices, name)
+        if callable(_matrix):
+            _matrix = _matrix(2 ** len(gate.target_qubits))
+        return self.cast(_matrix, dtype=_matrix.dtype)
 
-    def asmatrix_parametrized(self, gate):
+    def matrix_parametrized(self, gate):
         """Convert a parametrized gate to its matrix representation in the computational basis."""
         name = gate.__class__.__name__
-        return getattr(self.matrices, name)(*gate.parameters)
+        _matrix = getattr(self.matrices, name)(*gate.parameters)
+        return self.cast(_matrix, dtype=_matrix.dtype)
 
-    def asmatrix_fused(self, fgate):
+    def matrix_fused(self, fgate):
         rank = len(fgate.target_qubits)
-        matrix = np.eye(2**rank, dtype=self.dtype)
+        matrix = sparse.eye(2**rank)
+
         for gate in fgate.gates:
             # transfer gate matrix to numpy as it is more efficient for
             # small tensor calculations
-            gmatrix = gate.asmatrix(self)
+            # explicit to_numpy see https://github.com/qiboteam/qibo/issues/928
+            gmatrix = self.to_numpy(gate.matrix(self))
+            # add controls if controls were instantiated using
+            # the ``Gate.controlled_by`` method
+            num_controls = len(gate.control_qubits)
+            if num_controls > 0:
+                gmatrix = block_diag(
+                    np.eye(2 ** len(gate.qubits) - len(gmatrix)), gmatrix
+                )
             # Kronecker product with identity is needed to make the
             # original matrix have shape (2**rank x 2**rank)
-            eye = np.eye(2 ** (rank - len(gate.qubits)), dtype=self.dtype)
+            eye = np.eye(2 ** (rank - len(gate.qubits)))
             gmatrix = np.kron(gmatrix, eye)
             # Transpose the new matrix indices so that it targets the
             # target qubits of the original gate
@@ -128,34 +150,15 @@ class NumpyBackend(Backend):
             gmatrix = np.transpose(gmatrix, transpose_indices)
             gmatrix = np.reshape(gmatrix, original_shape)
             # fuse the individual gate matrix to the total ``FusedGate`` matrix
-            matrix = gmatrix @ matrix
-        return matrix
+            # we are using sparse matrices to improve perfomances
+            matrix = sparse.csr_matrix(gmatrix).dot(matrix)
 
-    def control_matrix(self, gate):
-        if len(gate.control_qubits) > 1:
-            raise_error(
-                NotImplementedError,
-                "Cannot calculate controlled "
-                "unitary for more than two "
-                "control qubits.",
-            )
-        matrix = gate.asmatrix(self)
-        shape = matrix.shape
-        if shape != (2, 2):
-            raise_error(
-                ValueError,
-                "Cannot use ``control_unitary`` method on "
-                "gate matrix of shape {}.".format(shape),
-            )
-        zeros = self.np.zeros((2, 2), dtype=self.dtype)
-        part1 = self.np.concatenate([self.np.eye(2, dtype=self.dtype), zeros], axis=0)
-        part2 = self.np.concatenate([zeros, matrix], axis=0)
-        return self.np.concatenate([part1, part2], axis=1)
+        return self.cast(matrix.toarray())
 
     def apply_gate(self, gate, state, nqubits):
         state = self.cast(state)
         state = self.np.reshape(state, nqubits * (2,))
-        matrix = gate.asmatrix(self)
+        matrix = gate.matrix(self)
         if gate.is_controlled_by:
             matrix = self.np.reshape(matrix, 2 * len(gate.target_qubits) * (2,))
             ncontrol = len(gate.control_qubits)
@@ -169,7 +172,7 @@ class NumpyBackend(Backend):
             updates = self.np.einsum(opstring, state[-1], matrix)
             # Concatenate the updated part of the state `updates` with the
             # part of of the state that remained unaffected `state[:-1]`.
-            state = self.np.concatenate([state[:-1], updates[self.np.newaxis]], axis=0)
+            state = self.np.concatenate([state[:-1], updates[None]], axis=0)
             state = self.np.reshape(state, nqubits * (2,))
             # Put qubit indices back to their proper places
             state = self.np.transpose(state, einsum_utils.reverse_order(order))
@@ -182,7 +185,7 @@ class NumpyBackend(Backend):
     def apply_gate_density_matrix(self, gate, state, nqubits):
         state = self.cast(state)
         state = self.np.reshape(state, 2 * nqubits * (2,))
-        matrix = gate.asmatrix(self)
+        matrix = gate.matrix(self)
         if gate.is_controlled_by:
             matrix = self.np.reshape(matrix, 2 * len(gate.target_qubits) * (2,))
             matrixc = self.np.conj(matrix)
@@ -211,11 +214,9 @@ class NumpyBackend(Backend):
 
             state00 = state[range(n - 1)]
             state00 = state00[:, range(n - 1)]
-            state01 = self.np.concatenate(
-                [state00, state01[:, self.np.newaxis]], axis=1
-            )
-            state10 = self.np.concatenate([state10, state11[self.np.newaxis]], axis=0)
-            state = self.np.concatenate([state01, state10[self.np.newaxis]], axis=0)
+            state01 = self.np.concatenate([state00, state01[:, None]], axis=1)
+            state10 = self.np.concatenate([state10, state11[None]], axis=0)
+            state = self.np.concatenate([state01, state10[None]], axis=0)
             state = self.np.reshape(state, 2 * nqubits * (2,))
             state = self.np.transpose(state, einsum_utils.reverse_order(order))
         else:
@@ -230,8 +231,8 @@ class NumpyBackend(Backend):
 
     def apply_gate_half_density_matrix(self, gate, state, nqubits):
         state = self.cast(state)
-        state = np.reshape(state, 2 * nqubits * (2,))
-        matrix = gate.asmatrix(self)
+        state = self.np.reshape(state, 2 * nqubits * (2,))
+        matrix = gate.matrix(self)
         if gate.is_controlled_by:  # pragma: no cover
             raise_error(
                 NotImplementedError,
@@ -240,17 +241,19 @@ class NumpyBackend(Backend):
                 "gates.",
             )
         else:
-            matrix = np.reshape(matrix, 2 * len(gate.qubits) * (2,))
+            matrix = self.np.reshape(matrix, 2 * len(gate.qubits) * (2,))
             left, _ = einsum_utils.apply_gate_density_matrix_string(
                 gate.qubits, nqubits
             )
-            state = np.einsum(left, state, matrix)
-        return np.reshape(state, 2 * (2**nqubits,))
+            state = self.np.einsum(left, state, matrix)
+        return self.np.reshape(state, 2 * (2**nqubits,))
 
     def apply_channel(self, channel, state, nqubits):
-        for coeff, gate in zip(channel.coefficients, channel.gates):
-            if self.np.random.random() < coeff:
-                state = self.apply_gate(gate, state, nqubits)
+        probabilities = channel.coefficients + (1 - np.sum(channel.coefficients),)
+        index = self.sample_shots(probabilities, 1)[0]
+        if index != len(channel.gates):
+            gate = channel.gates[index]
+            state = self.apply_gate(gate, state, nqubits)
         return state
 
     def apply_channel_density_matrix(self, channel, state, nqubits):
@@ -263,11 +266,12 @@ class NumpyBackend(Backend):
     def _append_zeros(self, state, qubits, results):
         """Helper method for collapse."""
         for q, r in zip(qubits, results):
-            state = self.np.expand_dims(state, axis=q)
-            if r:
-                state = self.np.concatenate([self.np.zeros_like(state), state], axis=q)
-            else:
-                state = self.np.concatenate([state, self.np.zeros_like(state)], axis=q)
+            state = self.np.expand_dims(state, q)
+            state = (
+                self.np.concatenate([self.np.zeros_like(state), state], q)
+                if r == 1
+                else self.np.concatenate([state, self.np.zeros_like(state)], q)
+            )
         return state
 
     def collapse_state(self, state, qubits, shot, nqubits, normalize=True):
@@ -310,17 +314,17 @@ class NumpyBackend(Backend):
         state = self.cast(state)
         shape = state.shape
         q = gate.target_qubits[0]
-        p0, p1 = gate.coefficients[:2]
+        p_0, p_1 = gate.init_kwargs["p_0"], gate.init_kwargs["p_1"]
         trace = self.partial_trace_density_matrix(state, (q,), nqubits)
         trace = self.np.reshape(trace, 2 * (nqubits - 1) * (2,))
         zero = self.zero_density_matrix(1)
-        zero = self.np.tensordot(trace, zero, axes=0)
+        zero = self.np.tensordot(trace, zero, 0)
         order = list(range(2 * nqubits - 2))
         order.insert(q, 2 * nqubits - 2)
         order.insert(q + nqubits, 2 * nqubits - 1)
         zero = self.np.reshape(self.np.transpose(zero, order), shape)
-        state = (1 - p0 - p1) * state + p0 * zero
-        return state + p1 * self.apply_gate_density_matrix(X(q), zero, nqubits)
+        state = (1 - p_0 - p_1) * state + p_0 * zero
+        return state + p_1 * self.apply_gate_density_matrix(X(q), zero, nqubits)
 
     def thermal_error_density_matrix(self, gate, state, nqubits):
         state = self.cast(state)
@@ -328,19 +332,82 @@ class NumpyBackend(Backend):
         state = self.apply_gate(gate, state.ravel(), 2 * nqubits)
         return self.np.reshape(state, shape)
 
-    def execute_circuit(
-        self, circuit, initial_state=None, nshots=None, return_array=False
-    ):
+    def depolarizing_error_density_matrix(self, gate, state, nqubits):
+        state = self.cast(state)
+        shape = state.shape
+        q = gate.target_qubits
+        lam = gate.init_kwargs["lam"]
+        trace = self.partial_trace_density_matrix(state, q, nqubits)
+        trace = self.np.reshape(trace, 2 * (nqubits - len(q)) * (2,))
+        identity = self.identity_density_matrix(len(q))
+        identity = self.np.reshape(identity, 2 * len(q) * (2,))
+        identity = self.np.tensordot(trace, identity, 0)
+        qubits = list(range(nqubits))
+        for j in q:
+            qubits.pop(qubits.index(j))
+        qubits.sort()
+        qubits += list(q)
+        qubit_1 = list(range(nqubits - len(q))) + list(
+            range(2 * (nqubits - len(q)), 2 * nqubits - len(q))
+        )
+        qubit_2 = list(range(nqubits - len(q), 2 * (nqubits - len(q)))) + list(
+            range(2 * nqubits - len(q), 2 * nqubits)
+        )
+        qs = [qubit_1, qubit_2]
+        order = []
+        for qj in qs:
+            qj = [qj[qubits.index(i)] for i in range(len(qubits))]
+            order += qj
+        identity = self.np.reshape(self.np.transpose(identity, order), shape)
+        state = (1 - lam) * state + lam * identity
+        return state
+
+    def execute_circuit(self, circuit, initial_state=None, nshots=1000):
+
+        if isinstance(initial_state, type(circuit)):
+            if not initial_state.density_matrix == circuit.density_matrix:
+                raise_error(
+                    ValueError,
+                    f"""Cannot set circuit with density_matrix {initial_state.density_matrix} as
+                      initial state for circuit with density_matrix {circuit.density_matrix}.""",
+                )
+            elif (
+                not initial_state.accelerators == circuit.accelerators
+            ):  # pragma: no cover
+                raise_error(
+                    ValueError,
+                    f"""Cannot set circuit with accelerators {initial_state.density_matrix} as
+                      initial state for circuit with accelerators {circuit.density_matrix}.""",
+                )
+            else:
+                return self.execute_circuit(initial_state + circuit, None, nshots)
+        elif initial_state is not None:
+            initial_state = self.cast(initial_state)
+            valid_shape = (
+                2 * (2**circuit.nqubits,)
+                if circuit.density_matrix
+                else (2**circuit.nqubits,)
+            )
+            if tuple(initial_state.shape) != valid_shape:
+                raise_error(
+                    ValueError,
+                    f"Given initial state has shape {initial_state.shape} instead of "
+                    f"the expected {valid_shape}.",
+                )
+
         if circuit.repeated_execution:
-            return self.execute_circuit_repeated(circuit, initial_state, nshots)
+            if circuit.measurements or circuit.has_collapse:
+                return self.execute_circuit_repeated(circuit, nshots, initial_state)
+            else:
+                raise RuntimeError(
+                    "Attempting to perform noisy simulation with `density_matrix=False` and no Measurement gate in the Circuit. If you wish to retrieve the statistics of the outcomes please include measurements in the circuit, otherwise set `density_matrix=True` to recover the final state."
+                )
 
         if circuit.accelerators:  # pragma: no cover
             return self.execute_distributed_circuit(circuit, initial_state, nshots)
 
         try:
             nqubits = circuit.nqubits
-            if isinstance(initial_state, CircuitResult):
-                initial_state = initial_state.state()
 
             if circuit.density_matrix:
                 if initial_state is None:
@@ -362,11 +429,27 @@ class NumpyBackend(Backend):
                 for gate in circuit.queue:
                     state = gate.apply(self, state, nqubits)
 
-            if return_array:
-                return state
+            if circuit.has_unitary_channel:
+                # here we necessarily have `density_matrix=True`, otherwise
+                # execute_circuit_repeated would have been called
+                if circuit.measurements:
+                    circuit._final_state = CircuitResult(
+                        state, circuit.measurements, backend=self, nshots=nshots
+                    )
+                    return circuit._final_state
+                else:
+                    circuit._final_state = QuantumState(state, backend=self)
+                    return circuit._final_state
+
             else:
-                circuit._final_state = CircuitResult(self, circuit, state, nshots)
-                return circuit._final_state
+                if circuit.measurements:
+                    circuit._final_state = CircuitResult(
+                        state, circuit.measurements, backend=self, nshots=nshots
+                    )
+                    return circuit._final_state
+                else:
+                    circuit._final_state = QuantumState(state, backend=self)
+                    return circuit._final_state
 
         except self.oom_error:
             raise_error(
@@ -376,12 +459,42 @@ class NumpyBackend(Backend):
                 "different one using ``qibo.set_device``.",
             )
 
-    def execute_circuit_repeated(self, circuit, initial_state=None, nshots=None):
-        if nshots is None:
-            nshots = 1
+    def execute_circuits(
+        self, circuits, initial_states=None, nshots=1000, processes=None
+    ):
+        from qibo.parallel import parallel_circuits_execution
 
-        results = []
+        return parallel_circuits_execution(
+            circuits, initial_states, nshots, processes, backend=self
+        )
+
+    def execute_circuit_repeated(self, circuit, nshots, initial_state=None):
+        """
+        Execute the circuit `nshots` times to retrieve probabilities, frequencies
+        and samples. Note that this method is called only if a unitary channel
+        is present in the circuit (i.e. noisy simulation) and `density_matrix=False`, or
+        if some collapsing measurement is performed.
+        """
+
+        if (
+            circuit.has_collapse
+            and not circuit.measurements
+            and not circuit.density_matrix
+        ):
+            raise RuntimeError(
+                "The circuit contains only collapsing measurements (`collapse=True`) but `density_matrix=False`. Please set `density_matrix=True` to retrieve the final state after execution."
+            )
+
+        results, final_states = [], []
         nqubits = circuit.nqubits
+
+        if not circuit.density_matrix:
+            samples = []
+            target_qubits = [
+                measurement.target_qubits for measurement in circuit.measurements
+            ]
+            target_qubits = sum(target_qubits, tuple())
+
         for _ in range(nshots):
             if circuit.density_matrix:
                 if initial_state is None:
@@ -393,13 +506,10 @@ class NumpyBackend(Backend):
                     if gate.symbolic_parameters:
                         gate.substitute_symbols()
                     state = gate.apply_density_matrix(self, state, nqubits)
-
             else:
                 if circuit.accelerators:  # pragma: no cover
                     # pylint: disable=E1111
-                    state = self.execute_distributed_circuit(
-                        circuit, initial_state, return_array=True
-                    )
+                    state = self.execute_distributed_circuit(circuit, initial_state)
                 else:
                     if initial_state is None:
                         state = self.zero_state(nqubits)
@@ -411,46 +521,52 @@ class NumpyBackend(Backend):
                             gate.substitute_symbols()
                         state = gate.apply(self, state, nqubits)
 
+            if circuit.density_matrix:
+                final_states.append(state)
             if circuit.measurements:
-                result = CircuitResult(self, circuit, state, 1)
-                results.append(result.samples()[0])
-            else:
-                results.append(state)
+                result = CircuitResult(
+                    state, circuit.measurements, backend=self, nshots=1
+                )
+                sample = result.samples()[0]
+                results.append(sample)
+                if not circuit.density_matrix:
+                    samples.append("".join([str(int(s)) for s in sample]))
+                for gate in circuit.measurements:
+                    gate.result.reset()
 
-        if circuit.measurements:
-            final_result = CircuitResult(self, circuit, state, nshots)
-            final_result._samples = self.aggregate_shots(results)
+        if circuit.density_matrix:  # this implies also it has_collapse
+            assert circuit.has_collapse
+            final_state = self.cast(np.mean(self.to_numpy(final_states), 0))
+            if circuit.measurements:
+                final_result = CircuitResult(
+                    final_state,
+                    circuit.measurements,
+                    backend=self,
+                    samples=self.aggregate_shots(results),
+                    nshots=nshots,
+                )
+            else:
+                final_result = QuantumState(final_state, backend=self)
             circuit._final_state = final_result
             return final_result
         else:
-            circuit._final_state = CircuitResult(self, circuit, results[-1], nshots)
-            return results
+            final_result = MeasurementOutcomes(
+                circuit.measurements,
+                backend=self,
+                samples=self.aggregate_shots(results),
+                nshots=nshots,
+            )
+            final_result._repeated_execution_frequencies = self.calculate_frequencies(
+                samples
+            )
+            circuit._final_state = final_result
+            return final_result
 
-    def execute_distributed_circuit(
-        self, circuit, initial_state=None, nshots=None, return_array=False
-    ):
+    def execute_distributed_circuit(self, circuit, initial_state=None, nshots=None):
         raise_error(
             NotImplementedError, f"{self} does not support distributed execution."
         )
 
-    def circuit_result_representation(self, result):
-        return result.symbolic()
-
-    def circuit_result_tensor(self, result):
-        return result.execution_result
-
-    def circuit_result_probabilities(self, result, qubits=None):
-        if qubits is None:
-            qubits = result.measurement_gate.qubits
-
-        state = self.circuit_result_tensor(result)
-        if result.density_matrix:
-            return self.calculate_probabilities_density_matrix(
-                state, qubits, result.nqubits
-            )
-        else:
-            return self.calculate_probabilities(state, qubits, result.nqubits)
-
     def calculate_symbolic(
         self, state, nqubits, decimals=5, cutoff=1e-10, max_terms=20
     ):
@@ -459,7 +575,7 @@ class NumpyBackend(Backend):
         for i in np.nonzero(state)[0]:
             b = bin(i)[2:].zfill(nqubits)
             if np.abs(state[i]) >= cutoff:
-                x = round(state[i], decimals)
+                x = np.round(state[i], decimals)
                 terms.append(f"{x}|{b}>")
             if len(terms) >= max_terms:
                 terms.append("...")
@@ -476,7 +592,7 @@ class NumpyBackend(Backend):
             bi = bin(i)[2:].zfill(nqubits)
             bj = bin(j)[2:].zfill(nqubits)
             if np.abs(state[i, j]) >= cutoff:
-                x = round(state[i, j], decimals)
+                x = np.round(state[i, j], decimals)
                 terms.append(f"{x}|{bi}><{bj}|")
             if len(terms) >= max_terms:
                 terms.append("...")
@@ -497,7 +613,7 @@ class NumpyBackend(Backend):
         rtype = self.np.real(state).dtype
         unmeasured_qubits = tuple(i for i in range(nqubits) if i not in qubits)
         state = self.np.reshape(self.np.abs(state) ** 2, nqubits * (2,))
-        probs = self.np.sum(state.astype(rtype), axis=unmeasured_qubits)
+        probs = self.np.sum(self.cast(state, dtype=rtype), axis=unmeasured_qubits)
         return self._order_probabilities(probs, qubits, nqubits).ravel()
 
     def calculate_probabilities_density_matrix(self, state, qubits, nqubits):
@@ -520,21 +636,27 @@ class NumpyBackend(Backend):
         )
 
     def aggregate_shots(self, shots):
-        return self.np.array(shots, dtype=shots[0].dtype)
+        return self.cast(shots, dtype=shots[0].dtype)
 
     def samples_to_binary(self, samples, nqubits):
-        qrange = self.np.arange(nqubits - 1, -1, -1, dtype="int32")
-        return self.np.mod(self.np.right_shift(samples[:, self.np.newaxis], qrange), 2)
+        ### This is faster just staying @ NumPy.
+        qrange = np.arange(nqubits - 1, -1, -1, dtype=np.int32)
+        samples = self.to_numpy(samples)
+        return np.mod(np.right_shift(samples[:, None], qrange), 2)
 
     def samples_to_decimal(self, samples, nqubits):
-        qrange = self.np.arange(nqubits - 1, -1, -1, dtype="int32")
-        qrange = (2**qrange)[:, self.np.newaxis]
-        return self.np.matmul(samples, qrange)[:, 0]
+        ### This is faster just staying @ NumPy.
+        qrange = np.arange(nqubits - 1, -1, -1, dtype=np.int32)
+        qrange = (2**qrange)[:, None]
+        samples = np.asarray(samples.tolist())
+        return np.matmul(samples, qrange)[:, 0]
 
     def calculate_frequencies(self, samples):
-        res, counts = self.np.unique(samples, return_counts=True)
-        res, counts = self.np.array(res), self.np.array(counts)
-        return collections.Counter({k: v for k, v in zip(res, counts)})
+        # Samples are a list of strings so there is no advantage in using other backends
+        res, counts = np.unique(samples, return_counts=True)
+        res = self.to_numpy(res).tolist()
+        counts = self.to_numpy(counts).tolist()
+        return collections.Counter(dict(zip(res, counts)))
 
     def update_frequencies(self, frequencies, probabilities, nsamples):
         samples = self.sample_shots(probabilities, nsamples)
@@ -546,28 +668,31 @@ class NumpyBackend(Backend):
         from qibo.config import SHOT_BATCH_SIZE
 
         nprobs = probabilities / self.np.sum(probabilities)
-        frequencies = self.np.zeros(len(nprobs), dtype="int64")
+        frequencies = self.np.zeros(len(nprobs), dtype=self.np.int64)
         for _ in range(nshots // SHOT_BATCH_SIZE):
             frequencies = self.update_frequencies(frequencies, nprobs, SHOT_BATCH_SIZE)
         frequencies = self.update_frequencies(
             frequencies, nprobs, nshots % SHOT_BATCH_SIZE
         )
-        return collections.Counter({i: f for i, f in enumerate(frequencies) if f > 0})
+        return collections.Counter(
+            {i: int(f) for i, f in enumerate(frequencies) if f > 0}
+        )
 
     def apply_bitflips(self, noiseless_samples, bitflip_probabilities):
-        fprobs = self.np.array(bitflip_probabilities, dtype="float64")
-        sprobs = self.np.random.random(noiseless_samples.shape)
-        flip0 = self.np.array(sprobs < fprobs[0], dtype=noiseless_samples.dtype)
-        flip1 = self.np.array(sprobs < fprobs[1], dtype=noiseless_samples.dtype)
-        noisy_samples = noiseless_samples + (1 - noiseless_samples) * flip0
-        noisy_samples = noisy_samples - noiseless_samples * flip1
+        noiseless_samples = self.cast(noiseless_samples, dtype=noiseless_samples.dtype)
+        fprobs = self.cast(bitflip_probabilities, dtype="float64")
+        sprobs = self.cast(np.random.random(noiseless_samples.shape), dtype="float64")
+        flip_0 = self.cast(sprobs < fprobs[0], dtype=noiseless_samples.dtype)
+        flip_1 = self.cast(sprobs < fprobs[1], dtype=noiseless_samples.dtype)
+        noisy_samples = noiseless_samples + (1 - noiseless_samples) * flip_0
+        noisy_samples = noisy_samples - noiseless_samples * flip_1
         return noisy_samples
 
     def partial_trace(self, state, qubits, nqubits):
         state = self.cast(state)
         state = self.np.reshape(state, nqubits * (2,))
         axes = 2 * [list(qubits)]
-        rho = self.np.tensordot(state, self.np.conj(state), axes=axes)
+        rho = self.np.tensordot(state, self.np.conj(state), axes)
         shape = 2 * (2 ** (nqubits - len(qubits)),)
         return self.np.reshape(rho, shape)
 
@@ -584,32 +709,21 @@ class NumpyBackend(Backend):
         state = self.np.reshape(state, shape)
         return self.np.einsum("abac->bc", state)
 
-    def entanglement_entropy(self, rho):
-        from qibo.config import EIGVAL_CUTOFF
-
-        # Diagonalize
-        eigvals = self.np.linalg.eigvalsh(rho).real
-        # Treating zero and negative eigenvalues
-        masked_eigvals = eigvals[eigvals > EIGVAL_CUTOFF]
-        spectrum = -1 * self.np.log(masked_eigvals)
-        entropy = self.np.sum(masked_eigvals * spectrum) / self.np.log(2.0)
-        return entropy, spectrum
-
-    def calculate_norm(self, state):
+    def calculate_norm(self, state, order=2):
         state = self.cast(state)
-        return self.np.sqrt(self.np.sum(self.np.abs(state) ** 2))
+        return self.np.linalg.norm(state, order)
 
-    def calculate_norm_density_matrix(self, state):
+    def calculate_norm_density_matrix(self, state, order="nuc"):
         state = self.cast(state)
-        return self.np.trace(state)
+        return self.np.linalg.norm(state, ord=order)
 
     def calculate_overlap(self, state1, state2):
-        state1 = self.cast(state1)
-        state2 = self.cast(state2)
-        return self.np.abs(self.np.sum(self.np.conj(state1) * state2))
+        return self.np.abs(self.np.sum(np.conj(self.cast(state1)) * self.cast(state2)))
 
     def calculate_overlap_density_matrix(self, state1, state2):
-        raise_error(NotImplementedError)
+        return self.np.trace(
+            self.np.matmul(self.np.conj(self.cast(state1)).T, self.cast(state2))
+        )
 
     def calculate_eigenvalues(self, matrix, k=6):
         if self.issparse(matrix):
@@ -637,44 +751,43 @@ class NumpyBackend(Backend):
             else:
                 from scipy.linalg import expm
             return expm(-1j * a * matrix)
-        else:
-            expd = self.np.diag(self.np.exp(-1j * a * eigenvalues))
-            ud = self.np.transpose(self.np.conj(eigenvectors))
-            return self.np.matmul(eigenvectors, self.np.matmul(expd, ud))
+        expd = self.np.diag(self.np.exp(-1j * a * eigenvalues))
+        ud = self.np.transpose(np.conj(eigenvectors))
+        return self.np.matmul(eigenvectors, self.np.matmul(expd, ud))
 
     def calculate_expectation_state(self, hamiltonian, state, normalize):
         statec = self.np.conj(state)
         hstate = hamiltonian @ state
         ev = self.np.real(self.np.sum(statec * hstate))
         if normalize:
-            norm = self.np.sum(self.np.square(self.np.abs(state)))
-            ev = ev / norm
+            ev /= self.np.sum(self.np.square(self.np.abs(state)))
         return ev
 
     def calculate_expectation_density_matrix(self, hamiltonian, state, normalize):
-        ev = self.np.real(self.np.trace(hamiltonian @ state))
+        ev = self.np.real(self.np.trace(self.cast(hamiltonian @ state)))
         if normalize:
             norm = self.np.real(self.np.trace(state))
-            ev = ev / norm
+            ev /= norm
         return ev
 
+    # TODO: remove this method
     def calculate_hamiltonian_matrix_product(self, matrix1, matrix2):
-        return self.np.dot(matrix1, matrix2)
+        return matrix1 @ matrix2
 
+    # TODO: remove this method
     def calculate_hamiltonian_state_product(self, matrix, state):
-        rank = len(tuple(state.shape))
-        state = self.cast(state)
-        if rank == 1:  # vector
-            return matrix.dot(state[:, np.newaxis])[:, 0]
-        elif rank == 2:  # matrix
-            return matrix.dot(state)
-        else:
+        if len(tuple(state.shape)) > 2:
             raise_error(
                 ValueError,
-                "Cannot multiply Hamiltonian with " "rank-{} tensor.".format(rank),
+                f"Cannot multiply Hamiltonian with rank-{len(tuple(state.shape))} tensor.",
             )
+        return matrix @ state
 
     def assert_allclose(self, value, target, rtol=1e-7, atol=0.0):
+        if isinstance(value, CircuitResult) or isinstance(value, QuantumState):
+            value = value.state()
+        if isinstance(target, CircuitResult) or isinstance(target, QuantumState):
+            target = target.state()
         value = self.to_numpy(value)
         target = self.to_numpy(target)
         np.testing.assert_allclose(value, target, rtol=rtol, atol=atol)
