@@ -14,151 +14,17 @@ class TensorflowMatrices(NumpyMatrices):
 
     def __init__(self, dtype):
         super().__init__(dtype)
-        import tensorflow as tf
-        import tensorflow.experimental.numpy as tnp  # pylint: disable=E0401
+        import tensorflow as tf  # pylint: disable=import-error
+        import tensorflow.experimental.numpy as tnp  # pylint: disable=import-error
 
         self.tf = tf
         self.np = tnp
 
-    def RX(self, theta):
-        cos = self.np.cos(theta / 2.0) + 0j
-        isin = -1j * self.np.sin(theta / 2.0)
-        return self.tf.cast([[cos, isin], [isin, cos]], dtype=self.dtype)
-
-    def RY(self, theta):
-        cos = self.np.cos(theta / 2.0) + 0j
-        sin = self.np.sin(theta / 2.0) + 0j
-        return self.tf.cast([[cos, -sin], [sin, cos]], dtype=self.dtype)
-
-    def RZ(self, theta):
-        phase = self.np.exp(0.5j * theta)
-        return self.tf.cast([[self.np.conj(phase), 0], [0, phase]], dtype=self.dtype)
-
-    def U1(self, theta):
-        phase = self.np.exp(1j * theta)
-        return self.tf.cast([[1, 0], [0, phase]], dtype=self.dtype)
-
-    def U2(self, phi, lam):
-        eplus = self.np.exp(1j * (phi + lam) / 2.0)
-        eminus = self.np.exp(1j * (phi - lam) / 2.0)
-        return self.tf.cast(
-            [[self.np.conj(eplus), -self.np.conj(eminus)], [eminus, eplus]],
-            dtype=self.dtype,
-        ) / self.np.sqrt(2)
-
-    def U3(self, theta, phi, lam):
-        cost = self.np.cos(theta / 2)
-        sint = self.np.sin(theta / 2)
-        eplus = self.np.exp(1j * (phi + lam) / 2.0)
-        eminus = self.np.exp(1j * (phi - lam) / 2.0)
-        return self.tf.cast(
-            [
-                [self.np.conj(eplus) * cost, -self.np.conj(eminus) * sint],
-                [eminus * sint, eplus * cost],
-            ],
-            dtype=self.dtype,
-        )
-
-    def CRX(self, theta):
-        r = self.RX(theta)
-        return self.tf.cast(
-            [
-                [1, 0, 0, 0],
-                [0, 1, 0, 0],
-                [0, 0, r[0, 0], r[0, 1]],
-                [0, 0, r[1, 0], r[1, 1]],
-            ],
-            dtype=self.dtype,
-        )
-
-    def CRY(self, theta):
-        r = self.RY(theta)
-        return self.tf.cast(
-            [
-                [1, 0, 0, 0],
-                [0, 1, 0, 0],
-                [0, 0, r[0, 0], r[0, 1]],
-                [0, 0, r[1, 0], r[1, 1]],
-            ],
-            dtype=self.dtype,
-        )
-
-    def CRZ(self, theta):
-        r = self.RZ(theta)
-        return self.tf.cast(
-            [
-                [1, 0, 0, 0],
-                [0, 1, 0, 0],
-                [0, 0, r[0, 0], r[0, 1]],
-                [0, 0, r[1, 0], r[1, 1]],
-            ],
-            dtype=self.dtype,
-        )
-
-    def CU1(self, theta):
-        r = self.U1(theta)
-        return self.tf.cast(
-            [
-                [1, 0, 0, 0],
-                [0, 1, 0, 0],
-                [0, 0, r[0, 0], r[0, 1]],
-                [0, 0, r[1, 0], r[1, 1]],
-            ],
-            dtype=self.dtype,
-        )
-
-    def CU2(self, phi, lam):
-        r = self.U2(phi, lam)
-        return self.tf.cast(
-            [
-                [1, 0, 0, 0],
-                [0, 1, 0, 0],
-                [0, 0, r[0, 0], r[0, 1]],
-                [0, 0, r[1, 0], r[1, 1]],
-            ],
-            dtype=self.dtype,
-        )
-
-    def CU3(self, theta, phi, lam):
-        r = self.U3(theta, phi, lam)
-        return self.tf.cast(
-            [
-                [1, 0, 0, 0],
-                [0, 1, 0, 0],
-                [0, 0, r[0, 0], r[0, 1]],
-                [0, 0, r[1, 0], r[1, 1]],
-            ],
-            dtype=self.dtype,
-        )
-
-    def fSim(self, theta, phi):
-        cost = self.np.cos(theta) + 0j
-        isint = -1j * self.np.sin(theta)
-        phase = self.np.exp(-1j * phi)
-        return self.tf.cast(
-            [
-                [1, 0, 0, 0],
-                [0, cost, isint, 0],
-                [0, isint, cost, 0],
-                [0, 0, 0, phase],
-            ],
-            dtype=self.dtype,
-        )
-
-    def GeneralizedfSim(self, u, phi):
-        phase = self.np.exp(-1j * phi)
-        return self.tf.cast(
-            [
-                [1, 0, 0, 0],
-                [0, u[0, 0], u[0, 1], 0],
-                [0, u[1, 0], u[1, 1], 0],
-                [0, 0, 0, phase],
-            ],
-            dtype=self.dtype,
-        )
+    def _cast(self, x, dtype):
+        return self.tf.cast(x, dtype=dtype)
 
     def Unitary(self, u):
-        return self.tf.cast(u, dtype=self.dtype)
+        return self._cast(u, dtype=self.dtype)
 
 
 class TensorflowBackend(NumpyBackend):
@@ -166,8 +32,12 @@ class TensorflowBackend(NumpyBackend):
         super().__init__()
         self.name = "tensorflow"
         os.environ["TF_CPP_MIN_LOG_LEVEL"] = str(TF_LOG_LEVEL)
-        import tensorflow as tf
-        import tensorflow.experimental.numpy as tnp  # pylint: disable=E0401
+
+        import tensorflow as tf  # pylint: disable=import-error
+        import tensorflow.experimental.numpy as tnp  # pylint: disable=import-error
+
+        if TF_LOG_LEVEL >= 2:
+            tf.get_logger().setLevel("ERROR")
 
         tnp.experimental_enable_numpy_behavior()
         self.tf = tf
@@ -181,7 +51,9 @@ class TensorflowBackend(NumpyBackend):
 
         self.matrices = TensorflowMatrices(self.dtype)
 
-        from tensorflow.python.framework import errors_impl  # pylint: disable=E0611
+        from tensorflow.python.framework import (  # pylint: disable=E0611,import-error
+            errors_impl,
+        )
 
         self.oom_error = errors_impl.ResourceExhaustedError
 
@@ -193,9 +65,7 @@ class TensorflowBackend(NumpyBackend):
         elif cpu_devices:
             self.device = cpu_devices[0].name
 
-        import psutil
-
-        self.nthreads = psutil.cpu_count(logical=True)
+        self.nthreads = 0
 
         self.tensor_types = (np.ndarray, tf.Tensor, tf.Variable)
 
@@ -242,27 +112,25 @@ class TensorflowBackend(NumpyBackend):
         state = self.tf.tensor_scatter_nd_update(state, idx, update)
         return state
 
-    def asmatrix(self, gate):
-        npmatrix = super().asmatrix(gate)
+    def matrix(self, gate):
+        npmatrix = super().matrix(gate)
         return self.tf.cast(npmatrix, dtype=self.dtype)
 
-    def asmatrix_parametrized(self, gate):
-        npmatrix = super().asmatrix_parametrized(gate)
+    def matrix_parametrized(self, gate):
+        npmatrix = super().matrix_parametrized(gate)
         return self.tf.cast(npmatrix, dtype=self.dtype)
 
-    def asmatrix_fused(self, gate):
-        npmatrix = super().asmatrix_fused(gate)
+    def matrix_fused(self, gate):
+        npmatrix = super().matrix_fused(gate)
         return self.tf.cast(npmatrix, dtype=self.dtype)
 
-    def execute_circuit(
-        self, circuit, initial_state=None, nshots=None, return_array=False
-    ):
+    def execute_circuit(self, circuit, initial_state=None, nshots=1000):
         with self.tf.device(self.device):
-            return super().execute_circuit(circuit, initial_state, nshots, return_array)
+            return super().execute_circuit(circuit, initial_state, nshots)
 
-    def execute_circuit_repeated(self, circuit, initial_state=None, nshots=None):
+    def execute_circuit_repeated(self, circuit, nshots, initial_state=None):
         with self.tf.device(self.device):
-            return super().execute_circuit_repeated(circuit, initial_state, nshots)
+            return super().execute_circuit_repeated(circuit, nshots, initial_state)
 
     def sample_shots(self, probabilities, nshots):
         # redefining this because ``tnp.random.choice`` is not available
@@ -281,7 +149,11 @@ class TensorflowBackend(NumpyBackend):
         # redefining this because ``tnp.unique`` is not available
         res, _, counts = self.tf.unique_with_counts(samples, out_idx="int64")
         res, counts = self.np.array(res), self.np.array(counts)
-        return collections.Counter({int(k): int(v) for k, v in zip(res, counts)})
+        if res.dtype == "string":
+            res = [r.numpy().decode("utf8") for r in res]
+        else:
+            res = [int(r) for r in res]
+        return collections.Counter({k: int(v) for k, v in zip(res, counts)})
 
     def update_frequencies(self, frequencies, probabilities, nsamples):
         # redefining this because ``tnp.unique`` and tensor update is not available
@@ -292,17 +164,15 @@ class TensorflowBackend(NumpyBackend):
         )
         return frequencies
 
-    def entanglement_entropy(self, rho):
-        # redefining this because ``tnp.linalg`` is not available
-        from qibo.config import EIGVAL_CUTOFF
+    def calculate_norm(self, state, order=2):
+        state = self.cast(state)
+        return self.tf.norm(state, ord=order)
 
-        # Diagonalize
-        eigvals = self.np.real(self.tf.linalg.eigvalsh(rho))
-        # Treating zero and negative eigenvalues
-        masked_eigvals = eigvals[eigvals > EIGVAL_CUTOFF]
-        spectrum = -1 * self.np.log(masked_eigvals)
-        entropy = self.np.sum(masked_eigvals * spectrum) / self.np.log(2.0)
-        return entropy, spectrum
+    def calculate_norm_density_matrix(self, state, order="nuc"):
+        state = self.cast(state)
+        if order == "nuc":
+            return self.np.trace(state)
+        return self.tf.norm(state, ord=order)
 
     def calculate_eigenvalues(self, matrix, k=6):
         return self.tf.linalg.eigvalsh(matrix)
@@ -333,16 +203,16 @@ class TensorflowBackend(NumpyBackend):
         else:
             raise_error(
                 ValueError,
-                "Cannot multiply Hamiltonian with " "rank-{} tensor.".format(rank),
+                f"Cannot multiply Hamiltonian with rank-{rank} tensor.",
             )
 
     def test_regressions(self, name):
         if name == "test_measurementresult_apply_bitflips":
             return [
-                [4, 0, 0, 1, 0, 2, 2, 4, 4, 0],
-                [4, 0, 0, 1, 0, 2, 2, 4, 4, 0],
-                [4, 0, 0, 1, 0, 0, 0, 4, 4, 0],
-                [4, 0, 0, 0, 0, 0, 0, 4, 4, 0],
+                [4, 0, 0, 1, 0, 0, 1, 0, 0, 0],
+                [0, 1, 1, 2, 1, 1, 4, 0, 0, 4],
+                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
+                [0, 0, 0, 0, 0, 4, 0, 0, 0, 4],
             ]
         elif name == "test_probabilistic_measurement":
             if "GPU" in self.device:  # pragma: no cover
@@ -357,6 +227,6 @@ class TensorflowBackend(NumpyBackend):
         elif name == "test_post_measurement_bitflips_on_circuit":
             return [
                 {5: 30},
-                {5: 16, 7: 10, 6: 2, 3: 1, 4: 1},
-                {3: 6, 5: 6, 7: 5, 2: 4, 4: 3, 0: 2, 1: 2, 6: 2},
+                {5: 12, 7: 6, 4: 6, 1: 5, 6: 1},
+                {3: 7, 6: 4, 2: 4, 7: 4, 0: 4, 5: 3, 4: 2, 1: 2},
             ]
