@@ -1,4 +1,4 @@
-from typing import List, Optional, Set, Union
+from typing import List, Optional, Union
 
 from qibo.config import raise_error
 
@@ -36,14 +36,13 @@ class Callback:
 
     def __getitem__(self, k):
         if not isinstance(k, (int, slice, list, tuple)):
-            raise_error(IndexError, "Unrecognized type for index {}.".format(k))
+            raise_error(IndexError, f"Unrecognized type for index {k}.")
 
         if isinstance(k, int) and k >= len(self._results):
             raise_error(
                 IndexError,
-                "Attempting to access callbacks {} run but "
-                "the callback has been used in {} executions."
-                "".format(k, len(self._results)),
+                f"Attempting to access callbacks {k} run but "
+                + f"the callback has been used in {len(self._results)} executions.",
             )
 
         return self._results[k]
@@ -96,17 +95,21 @@ class EntanglementEntropy(Callback):
     """
 
     def __init__(
-        self, partition: Optional[List[int]] = None, compute_spectrum: bool = False
+        self,
+        partition: Optional[List[int]] = None,
+        compute_spectrum: bool = False,
+        base: float = 2,
+        check_hermitian: bool = False,
     ):
         super().__init__()
         self.partition = partition
         self.compute_spectrum = compute_spectrum
+        self.base = base
+        self.check_hermitian = check_hermitian
         self.spectrum = list()
 
     @Callback.nqubits.setter
     def nqubits(self, n: int):
-        from qibo import gates
-
         if self._nqubits is not None and self._nqubits != n:
             raise_error(
                 RuntimeError,
@@ -115,26 +118,29 @@ class EntanglementEntropy(Callback):
         self._nqubits = n
         if self.partition is None:
             self.partition = list(range(n // 2 + n % 2))
-        if len(self.partition) <= self.nqubits // 2:
+        if len(self.partition) <= self._nqubits // 2:
             self.partition = [
-                i for i in range(self.nqubits) if i not in set(self.partition)
+                i for i in range(self._nqubits) if i not in set(self.partition)
             ]
 
     def apply(self, backend, state):
-        rho = backend.partial_trace(state, self.partition, self.nqubits)
-        entropy, spectrum = backend.entanglement_entropy(rho)
+        from qibo.quantum_info.entropies import entanglement_entropy
+
+        entropy, spectrum = entanglement_entropy(
+            state,
+            bipartition=self.partition,
+            base=self.base,
+            check_hermitian=self.check_hermitian,
+            return_spectrum=True,
+            backend=backend,
+        )
         self.append(entropy)
         if self.compute_spectrum:
             self.spectrum.append(spectrum)
         return entropy
 
     def apply_density_matrix(self, backend, state):
-        rho = backend.partial_trace_density_matrix(state, self.partition, self.nqubits)
-        entropy, spectrum = backend.entanglement_entropy(rho)
-        self.append(entropy)
-        if self.compute_spectrum:
-            self.spectrum.append(spectrum)
-        return entropy
+        return self.apply(backend, state)
 
 
 class State(Callback):
@@ -175,10 +181,14 @@ class Norm(Callback):
     """
 
     def apply(self, backend, state):
-        return backend.calculate_norm(state)
+        norm = backend.calculate_norm(state)
+        self.append(norm)
+        return norm
 
     def apply_density_matrix(self, backend, state):
-        return backend.calculate_norm_density_matrix(state)
+        norm = backend.calculate_norm_density_matrix(state)
+        self.append(norm)
+        return norm
 
 
 class Overlap(Callback):
@@ -200,10 +210,14 @@ class Overlap(Callback):
         self.state = state
 
     def apply(self, backend, state):
-        return backend.calculate_overlap(self.state, state)
+        overlap = backend.calculate_overlap(self.state, state)
+        self.append(overlap)
+        return overlap
 
     def apply_density_matrix(self, backend, state):
-        return backend.calculate_overlap_density_matrix(self.state, state)
+        overlap = backend.calculate_overlap_density_matrix(self.state, state)
+        self.append(overlap)
+        return overlap
 
 
 class Energy(Callback):
@@ -229,11 +243,15 @@ class Energy(Callback):
 
     def apply(self, backend, state):
         assert type(self.hamiltonian.backend) == type(backend)
-        return self.hamiltonian.expectation(state)
+        expectation = self.hamiltonian.expectation(state)
+        self.append(expectation)
+        return expectation
 
     def apply_density_matrix(self, backend, state):
         assert type(self.hamiltonian.backend) == type(backend)
-        return self.hamiltonian.expectation(state)
+        expectation = self.hamiltonian.expectation(state)
+        self.append(expectation)
+        return expectation
 
 
 class Gap(Callback):
@@ -289,13 +307,10 @@ class Gap(Callback):
         if not isinstance(mode, (int, str)):
             raise_error(
                 TypeError,
-                "Gap callback mode should be integer or "
-                "string but is {}.".format(type(mode)),
+                f"Gap callback mode should be integer or string but is {type(mode)}.",
             )
         elif isinstance(mode, str) and mode != "gap":
-            raise_error(
-                ValueError, "Unsupported mode {} for gap callback." "".format(mode)
-            )
+            raise_error(ValueError, f"Unsupported mode {mode} for gap callback.")
         self.mode = mode
         self.check_degenerate = check_degenerate
         self.evolution = None
@@ -314,12 +329,16 @@ class Gap(Callback):
         hamiltonian.eigenvectors()
         eigvals = hamiltonian.eigenvalues()
         if isinstance(self.mode, int):
-            return backend.np.real(eigvals[self.mode])
+            gap = backend.np.real(eigvals[self.mode])
+            self.append(gap)
+            return gap
 
         # case: self.mode == "gap"
         excited = 1
         gap = backend.np.real(eigvals[excited] - eigvals[0])
+
         if not self.check_degenerate:
+            self.append(gap)
             return gap
 
         while backend.np.less(gap, EIGVAL_CUTOFF):
@@ -327,9 +346,9 @@ class Gap(Callback):
             excited += 1
         if excited > 1:
             log.warning(
-                "The Hamiltonian is degenerate. Using eigenvalue {} "
-                "to calculate gap.".format(excited)
+                f"The Hamiltonian is degenerate. Using eigenvalue {excited} to calculate gap."
             )
+        self.append(gap)
         return gap
 
     def apply_density_matrix(self, backend, state):
