@@ -1,5 +1,12 @@
+"""Module defining Hamiltonian classes."""
+
+from itertools import chain
+from typing import Optional
+
+import numpy as np
 import sympy
 
+from qibo.backends import PyTorchBackend
 from qibo.config import EINSUM_CHARS, log, raise_error
 from qibo.hamiltonians.abstract import AbstractHamiltonian
 from qibo.symbols import Z
@@ -18,12 +25,9 @@ class Hamiltonian(AbstractHamiltonian):
     """
 
     def __init__(self, nqubits, matrix=None, backend=None):
-        if backend is None:  # pragma: no cover
-            from qibo.backends import GlobalBackend
+        from qibo.backends import _check_backend
 
-            self.backend = GlobalBackend()
-        else:
-            self.backend = backend
+        self.backend = _check_backend(backend)
 
         if not (
             isinstance(matrix, self.backend.tensor_types)
@@ -31,9 +35,7 @@ class Hamiltonian(AbstractHamiltonian):
         ):
             raise_error(
                 TypeError,
-                "Matrix of invalid type {} given during "
-                "Hamiltonian initialization"
-                "".format(type(matrix)),
+                f"Matrix of invalid type {type(matrix)} given during Hamiltonian initialization",
             )
         matrix = self.backend.cast(matrix)
 
@@ -54,22 +56,22 @@ class Hamiltonian(AbstractHamiltonian):
         return self._matrix
 
     @matrix.setter
-    def matrix(self, m):
-        shape = tuple(m.shape)
+    def matrix(self, mat):
+        shape = tuple(mat.shape)
         if shape != 2 * (2**self.nqubits,):
             raise_error(
                 ValueError,
-                "The Hamiltonian is defined for {} qubits "
-                "while the given matrix has shape {}."
-                "".format(self.nqubits, shape),
+                f"The Hamiltonian is defined for {self.nqubits} qubits "
+                + f"while the given matrix has shape {shape}.",
             )
-        self._matrix = m
+        self._matrix = mat
 
     @classmethod
     def from_symbolic(cls, symbolic_hamiltonian, symbol_map, backend=None):
         """Creates a ``Hamiltonian`` from a symbolic Hamiltonian.
 
-        We refer to the :ref:`How to define custom Hamiltonians using symbols? <symbolicham-example>`
+        We refer to the
+        :ref:`How to define custom Hamiltonians using symbols? <symbolicham-example>`
         example for more details.
 
         Args:
@@ -88,7 +90,9 @@ class Hamiltonian(AbstractHamiltonian):
             "deprecated. Please use `SymbolicHamiltonian` and Qibo symbols "
             "to construct Hamiltonians using symbols."
         )
-        return SymbolicHamiltonian(symbolic_hamiltonian, symbol_map, backend=backend)
+        return SymbolicHamiltonian(
+            symbolic_hamiltonian, symbol_map=symbol_map, backend=backend
+        )
 
     def eigenvalues(self, k=6):
         if self._eigenvalues is None:
@@ -112,31 +116,29 @@ class Hamiltonian(AbstractHamiltonian):
 
     def expectation(self, state, normalize=False):
         if isinstance(state, self.backend.tensor_types):
+            state = self.backend.cast(state)
             shape = tuple(state.shape)
             if len(shape) == 1:  # state vector
                 return self.backend.calculate_expectation_state(self, state, normalize)
-            elif len(shape) == 2:  # density matrix
+
+            if len(shape) == 2:  # density matrix
                 return self.backend.calculate_expectation_density_matrix(
                     self, state, normalize
                 )
-            else:
-                raise_error(
-                    ValueError,
-                    "Cannot calculate Hamiltonian "
-                    "expectation value for state of shape "
-                    "{}.".format(shape),
-                )
-        else:
+
             raise_error(
-                TypeError,
-                "Cannot calculate Hamiltonian expectation "
-                "value for state of type {}."
-                "".format(type(state)),
+                ValueError,
+                "Cannot calculate Hamiltonian expectation value "
+                + f"for state of shape {shape}",
             )
 
-    def expectation_from_samples(self, freq, qubit_map=None):
-        import numpy as np
+        raise_error(
+            TypeError,
+            "Cannot calculate Hamiltonian expectation "
+            + f"value for state of type {type(state)}",
+        )
 
+    def expectation_from_samples(self, freq, qubit_map=None):
         obs = self.matrix
         if np.count_nonzero(obs - np.diag(np.diagonal(obs))) != 0:
             raise_error(NotImplementedError, "Observable is not diagonal.")
@@ -145,25 +147,48 @@ class Hamiltonian(AbstractHamiltonian):
             qubit_map = list(range(int(np.log2(len(obs)))))
         counts = np.array(list(freq.values())) / sum(freq.values())
         expval = 0
-        kl = len(qubit_map)
+        size = len(qubit_map)
         for j, k in enumerate(keys):
             index = 0
             for i in qubit_map:
-                index += int(k[qubit_map.index(i)]) * 2 ** (kl - 1 - i)
+                index += int(k[qubit_map.index(i)]) * 2 ** (size - 1 - i)
             expval += obs[index, index] * counts[j]
-        return expval
+        return np.real(expval)
 
-    def eye(self, n=None):
-        if n is None:
-            n = int(self.matrix.shape[0])
-        return self.backend.cast(self.backend.matrices.I(n), dtype=self.matrix.dtype)
+    def eye(self, dim: Optional[int] = None):
+        """Generate Identity matrix with dimension ``dim``"""
+        if dim is None:
+            dim = int(self.matrix.shape[0])
+        return self.backend.cast(self.backend.matrices.I(dim), dtype=self.matrix.dtype)
+
+    def energy_fluctuation(self, state):
+        """
+        Evaluate energy fluctuation:
+
+        .. math::
+            \\Xi_{k}(\\mu) = \\sqrt{\\langle\\mu|\\hat{H}^2|\\mu\\rangle - \\langle\\mu|\\hat{H}|\\mu\\rangle^2} \\,
+
+        for a given state :math:`|\\mu\\rangle`.
+
+        Args:
+            state (np.ndarray): quantum state to be used to compute the energy fluctuation.
+
+        Return:
+            Energy fluctuation value (float).
+        """
+        state = self.backend.cast(state)
+        energy = self.expectation(state)
+        h = self.matrix
+        h2 = Hamiltonian(nqubits=self.nqubits, matrix=h @ h, backend=self.backend)
+        average_h2 = h2.expectation(state, normalize=True)
+        return np.sqrt(np.abs(average_h2 - energy**2))
 
     def __add__(self, o):
         if isinstance(o, self.__class__):
             if self.nqubits != o.nqubits:
                 raise_error(
                     RuntimeError,
-                    "Only hamiltonians with the same " "number of qubits can be added.",
+                    "Only hamiltonians with the same number of qubits can be added.",
                 )
             new_matrix = self.matrix + o.matrix
         elif isinstance(o, self.backend.numeric_types):
@@ -171,7 +196,7 @@ class Hamiltonian(AbstractHamiltonian):
         else:
             raise_error(
                 NotImplementedError,
-                "Hamiltonian addition to {} not " "implemented.".format(type(o)),
+                f"Hamiltonian addition to {type(o)} not implemented.",
             )
         return self.__class__(self.nqubits, new_matrix, backend=self.backend)
 
@@ -180,8 +205,7 @@ class Hamiltonian(AbstractHamiltonian):
             if self.nqubits != o.nqubits:
                 raise_error(
                     RuntimeError,
-                    "Only hamiltonians with the same "
-                    "number of qubits can be subtracted.",
+                    "Only hamiltonians with the same number of qubits can be subtracted.",
                 )
             new_matrix = self.matrix - o.matrix
         elif isinstance(o, self.backend.numeric_types):
@@ -189,7 +213,7 @@ class Hamiltonian(AbstractHamiltonian):
         else:
             raise_error(
                 NotImplementedError,
-                "Hamiltonian subtraction to {} " "not implemented.".format(type(o)),
+                f"Hamiltonian subtraction to {type(o)} not implemented.",
             )
         return self.__class__(self.nqubits, new_matrix, backend=self.backend)
 
@@ -199,7 +223,7 @@ class Hamiltonian(AbstractHamiltonian):
             if self.nqubits != o.nqubits:
                 raise_error(
                     RuntimeError,
-                    "Only hamiltonians with the same " "number of qubits can be added.",
+                    "Only hamiltonians with the same number of qubits can be added.",
                 )
             new_matrix = o.matrix - self.matrix
         elif isinstance(o, self.backend.numeric_types):
@@ -207,7 +231,7 @@ class Hamiltonian(AbstractHamiltonian):
         else:
             raise_error(
                 NotImplementedError,
-                "Hamiltonian subtraction to {} " "not implemented.".format(type(o)),
+                f"Hamiltonian subtraction to {type(o)} not implemented.",
             )
         return self.__class__(self.nqubits, new_matrix, backend=self.backend)
 
@@ -217,15 +241,17 @@ class Hamiltonian(AbstractHamiltonian):
         elif not isinstance(o, self.backend.numeric_types):
             raise_error(
                 NotImplementedError,
-                "Hamiltonian multiplication to {} " "not implemented.".format(type(o)),
+                f"Hamiltonian multiplication to {type(o)} not implemented.",
             )
         new_matrix = self.matrix * o
         r = self.__class__(self.nqubits, new_matrix, backend=self.backend)
+        o = self.backend.cast(o)
         if self._eigenvalues is not None:
             if self.backend.np.real(o) >= 0:  # TODO: check for side effects K.qnp
                 r._eigenvalues = o * self._eigenvalues
             elif not self.backend.issparse(self.matrix):
-                r._eigenvalues = o * self._eigenvalues[::-1]
+                axis = (0,) if isinstance(self.backend, PyTorchBackend) else 0
+                r._eigenvalues = o * self.backend.np.flip(self._eigenvalues, axis)
         if self._eigenvectors is not None:
             if self.backend.np.real(o) > 0:  # TODO: see above
                 r._eigenvectors = self._eigenvectors
@@ -240,53 +266,13 @@ class Hamiltonian(AbstractHamiltonian):
             )
             return self.__class__(self.nqubits, matrix, backend=self.backend)
 
-        elif isinstance(o, self.backend.tensor_types):
+        if isinstance(o, self.backend.tensor_types):
             return self.backend.calculate_hamiltonian_state_product(self.matrix, o)
 
-        else:
-            raise_error(
-                NotImplementedError,
-                "Hamiltonian matmul to {} not " "implemented.".format(type(o)),
-            )
-
-
-class TrotterCircuit:
-    """Object that caches the Trotterized evolution circuit.
-
-    This object holds a reference to the circuit models and updates its
-    parameters if a different time step ``dt`` is given without recreating
-    every gate from scratch.
-
-    Args:
-        groups (list): List of :class:`qibo.core.terms.TermGroup` objects that
-            correspond to the Trotter groups of terms in the time evolution
-            exponential operator.
-        dt (float): Time step for the Trotterization.
-        nqubits (int): Number of qubits in the system that evolves.
-        accelerators (dict): Dictionary with accelerators for distributed
-            circuits.
-    """
-
-    def __init__(self, groups, dt, nqubits, accelerators):
-        from itertools import chain
-
-        from qibo.models import Circuit
-
-        self.gates = {}
-        self.dt = dt
-        self.circuit = Circuit(nqubits, accelerators=accelerators)
-        for group in chain(groups, groups[::-1]):
-            gate = group.term.expgate(dt / 2.0)
-            self.gates[gate] = group
-            self.circuit.add(gate)
-
-    def set(self, dt):
-        if self.dt != dt:
-            params = {
-                gate: group.term.exp(dt / 2.0) for gate, group in self.gates.items()
-            }
-            self.dt = dt
-            self.circuit.set_parameters(params)
+        raise_error(
+            NotImplementedError,
+            f"Hamiltonian matmul to {type(o)} not implemented.",
+        )
 
 
 class SymbolicHamiltonian(AbstractHamiltonian):
@@ -318,7 +304,7 @@ class SymbolicHamiltonian(AbstractHamiltonian):
             :meth:`qibo.hamiltonians.models.MaxCut` Hamiltonian.
     """
 
-    def __init__(self, form=None, symbol_map={}, backend=None):
+    def __init__(self, form=None, nqubits=None, symbol_map={}, backend=None):
         super().__init__()
         self._form = None
         self._terms = None
@@ -327,18 +313,19 @@ class SymbolicHamiltonian(AbstractHamiltonian):
         self.symbol_map = symbol_map
         # if a symbol in the given form is not a Qibo symbol it must be
         # included in the ``symbol_map``
-        self.trotter_circuit = None
-        from qibo.symbols import Symbol
+
+        from qibo.symbols import Symbol  # pylint: disable=import-outside-toplevel
 
         self._qiboSymbol = Symbol  # also used in ``self._get_symbol_matrix``
-        if backend is None:  # pragma: no cover
-            from qibo.backends import GlobalBackend
 
-            self.backend = GlobalBackend()
-        else:
-            self.backend = backend
+        from qibo.backends import _check_backend
+
+        self.backend = _check_backend(backend)
+
         if form is not None:
             self.form = form
+        if nqubits is not None:
+            self.nqubits = nqubits
 
     @property
     def dense(self):
@@ -369,8 +356,7 @@ class SymbolicHamiltonian(AbstractHamiltonian):
         if not isinstance(form, sympy.Expr):
             raise_error(
                 TypeError,
-                "Symbolic Hamiltonian should be a ``sympy`` "
-                "expression but is {}.".format(type(form)),
+                f"Symbolic Hamiltonian should be a ``sympy`` expression but is {type(form)}.",
             )
         # Calculate number of qubits in the system described by the given
         # Hamiltonian formula
@@ -380,9 +366,7 @@ class SymbolicHamiltonian(AbstractHamiltonian):
                 q = symbol.target_qubit
             elif isinstance(symbol, sympy.Expr):
                 if symbol not in self.symbol_map:
-                    raise_error(
-                        ValueError, "Symbol {} is not in symbol " "map.".format(symbol)
-                    )
+                    raise_error(ValueError, f"Symbol {symbol} is not in symbol map.")
                 q, matrix = self.symbol_map.get(symbol)
                 if not isinstance(matrix, self.backend.tensor_types):
                     # ignore symbols that do not correspond to quantum operators
@@ -396,10 +380,14 @@ class SymbolicHamiltonian(AbstractHamiltonian):
 
     @property
     def terms(self):
-        """List of :class:`qibo.core.terms.HamiltonianTerm` objects of which the Hamiltonian is a sum of."""
+        """List of :class:`qibo.core.terms.HamiltonianTerm` objects
+        of which the Hamiltonian is a sum of.
+        """
         if self._terms is None:
             # Calculate terms based on ``self.form``
-            from qibo.hamiltonians.terms import SymbolicTerm
+            from qibo.hamiltonians.terms import (  # pylint: disable=import-outside-toplevel
+                SymbolicTerm,
+            )
 
             form = sympy.expand(self.form)
             terms = []
@@ -409,9 +397,6 @@ class SymbolicHamiltonian(AbstractHamiltonian):
                     terms.append(term)
                 else:
                     self.constant += term.coefficient
-            assert (
-                self.nqubits == max(q for term in terms for q in term.target_qubits) + 1
-            )
             self._terms = terms
         return self._terms
 
@@ -454,8 +439,6 @@ class SymbolicHamiltonian(AbstractHamiltonian):
             Numerical matrix corresponding to the given expression as a numpy
             array of size ``(2 ** self.nqubits, 2 ** self.nqubits).
         """
-        from numpy import eye
-
         if isinstance(term, sympy.Add):
             # symbolic op for addition
             result = sum(
@@ -492,7 +475,7 @@ class SymbolicHamiltonian(AbstractHamiltonian):
                 if not isinstance(matrix, self.backend.tensor_types):
                     # symbols that do not correspond to quantum operators
                     # for example parameters in the MaxCut Hamiltonian
-                    result = complex(matrix) * eye(2**self.nqubits)
+                    result = complex(matrix) * np.eye(2**self.nqubits)
                 else:
                     # if we do not have a Qibo symbol we construct one and use
                     # :meth:`qibo.core.terms.SymbolicTerm.full_matrix`.
@@ -502,13 +485,12 @@ class SymbolicHamiltonian(AbstractHamiltonian):
             # if the term is number we should return in the form of identity
             # matrix because in expressions like `1 + Z`, `1` is not correspond
             # to the float 1 but the identity operator (matrix)
-            result = complex(term) * eye(2**self.nqubits)
+            result = complex(term) * np.eye(2**self.nqubits)
 
         else:
             raise_error(
                 TypeError,
-                "Cannot calculate matrix for symbolic term "
-                "of type {}.".format(type(term)),
+                f"Cannot calculate matrix for symbolic term of type {type(term)}.",
             )
 
         return result
@@ -521,11 +503,9 @@ class SymbolicHamiltonian(AbstractHamiltonian):
         return Hamiltonian(self.nqubits, matrix, backend=self.backend)
 
     def _calculate_dense_from_terms(self):
-        """Calculates equivalent :class:`qibo.core.hamiltonians.Hamiltonian` using the term representation."""
-        from itertools import chain
-
-        import numpy as np
-
+        """Calculates equivalent :class:`qibo.core.hamiltonians.Hamiltonian`
+        using the term representation.
+        """
         if 2 * self.nqubits > len(EINSUM_CHARS):  # pragma: no cover
             # case not tested because it only happens in large examples
             raise_error(NotImplementedError, "Not enough einsum characters.")
@@ -555,13 +535,11 @@ class SymbolicHamiltonian(AbstractHamiltonian):
         return Hamiltonian.expectation(self, state, normalize)
 
     def expectation_from_samples(self, freq, qubit_map=None):
-        import numpy as np
-
         terms = self.terms
         for term in terms:
             # pylint: disable=E1101
             for factor in term.factors:
-                if isinstance(factor, Z) == False:
+                if not isinstance(factor, Z):
                     raise_error(
                         NotImplementedError, "Observable is not a Z Pauli string."
                     )
@@ -596,7 +574,7 @@ class SymbolicHamiltonian(AbstractHamiltonian):
             if self.nqubits != o.nqubits:
                 raise_error(
                     RuntimeError,
-                    "Only hamiltonians with the same " "number of qubits can be added.",
+                    "Only hamiltonians with the same number of qubits can be added.",
                 )
             new_ham = self.__class__(
                 symbol_map=dict(self.symbol_map), backend=self.backend
@@ -625,8 +603,7 @@ class SymbolicHamiltonian(AbstractHamiltonian):
         else:
             raise_error(
                 NotImplementedError,
-                "SymbolicHamiltonian addition to {} not "
-                "implemented.".format(type(o)),
+                f"SymbolicHamiltonian addition to {type(o)} not implemented.",
             )
         return new_ham
 
@@ -635,8 +612,7 @@ class SymbolicHamiltonian(AbstractHamiltonian):
             if self.nqubits != o.nqubits:
                 raise_error(
                     RuntimeError,
-                    "Only hamiltonians with the same "
-                    "number of qubits can be subtracted.",
+                    "Only hamiltonians with the same number of qubits can be subtracted.",
                 )
             new_ham = self.__class__(
                 symbol_map=dict(self.symbol_map), backend=self.backend
@@ -665,7 +641,7 @@ class SymbolicHamiltonian(AbstractHamiltonian):
         else:
             raise_error(
                 NotImplementedError,
-                "Hamiltonian subtraction to {} " "not implemented.".format(type(o)),
+                f"Hamiltonian subtraction to {type(o)} " "not implemented.",
             )
         return new_ham
 
@@ -684,18 +660,15 @@ class SymbolicHamiltonian(AbstractHamiltonian):
         else:
             raise_error(
                 NotImplementedError,
-                "Hamiltonian subtraction to {} " "not implemented.".format(type(o)),
+                f"Hamiltonian subtraction to {type(o)} not implemented.",
             )
         return new_ham
 
     def __mul__(self, o):
-        if not (
-            isinstance(o, self.backend.numeric_types)
-            or isinstance(o, self.backend.tensor_types)
-        ):
+        if not isinstance(o, (self.backend.numeric_types, self.backend.tensor_types)):
             raise_error(
                 NotImplementedError,
-                "Hamiltonian multiplication to {} " "not implemented.".format(type(o)),
+                f"Hamiltonian multiplication to {type(o)} not implemented.",
             )
         o = complex(o)
         new_ham = self.__class__(symbol_map=dict(self.symbol_map), backend=self.backend)
@@ -745,32 +718,47 @@ class SymbolicHamiltonian(AbstractHamiltonian):
 
         if isinstance(o, self.backend.tensor_types):
             rank = len(tuple(o.shape))
-            if rank == 1:  # state vector
-                return self.apply_gates(o)
-            elif rank == 2:  # density matrix
-                return self.apply_gates(o, density_matrix=True)
-            else:
+            if rank not in (1, 2):
                 raise_error(
                     NotImplementedError,
-                    "Cannot multiply Hamiltonian with " "rank-{} tensor.".format(rank),
+                    f"Cannot multiply Hamiltonian with rank-{rank} tensor.",
                 )
+            state_qubits = int(np.log2(int(o.shape[0])))
+            if state_qubits != self.nqubits:
+                raise_error(
+                    ValueError,
+                    f"Cannot multiply Hamiltonian on {self.nqubits} qubits to "
+                    + f"state of {state_qubits} qubits.",
+                )
+            if rank == 1:  # state vector
+                return self.apply_gates(o)
+
+            return self.apply_gates(o, density_matrix=True)
 
         raise_error(
             NotImplementedError,
-            "Hamiltonian matmul to {} not " "implemented.".format(type(o)),
+            f"Hamiltonian matmul to {type(o)} not implemented.",
         )
 
     def circuit(self, dt, accelerators=None):
-        """Circuit that implements a Trotter step of this Hamiltonian for a given time step ``dt``."""
-        if self.trotter_circuit is None:
-            from qibo.hamiltonians.terms import TermGroup
+        """Circuit that implements a Trotter step of this Hamiltonian.
 
-            groups = TermGroup.from_terms(self.terms)
-            self.trotter_circuit = TrotterCircuit(
-                groups, dt, self.nqubits, accelerators
-            )
-        self.trotter_circuit.set(dt)
-        return self.trotter_circuit.circuit
+        Args:
+            dt (float): Time step used for Trotterization.
+            accelerators (dict): Dictionary with accelerators for distributed circuits.
+        """
+        from qibo import Circuit  # pylint: disable=import-outside-toplevel
+        from qibo.hamiltonians.terms import (  # pylint: disable=import-outside-toplevel
+            TermGroup,
+        )
+
+        groups = TermGroup.from_terms(self.terms)
+        circuit = Circuit(self.nqubits, accelerators=accelerators)
+        circuit.add(
+            group.term.expgate(dt / 2.0) for group in chain(groups, groups[::-1])
+        )
+
+        return circuit
 
 
 class TrotterHamiltonian:
@@ -780,8 +768,8 @@ class TrotterHamiltonian:
         raise_error(
             NotImplementedError,
             "`TrotterHamiltonian` is substituted by `SymbolicHamiltonian` "
-            "and is no longer supported. Please check the documentation "
-            "of `SymbolicHamiltonian` for more details.",
+            + "and is no longer supported. Please check the documentation "
+            + "of `SymbolicHamiltonian` for more details.",
         )
 
     @classmethod
