@@ -1,15 +1,14 @@
 from qibo.models import hep, tsp
 from qibo.models.circuit import Circuit
-from qibo.models.error_mitigation import (
-    CDR,
-    ZNE,
-    get_gammas,
-    get_noisy_circuit,
-    sample_training_circuit,
-    vnCDR,
+from qibo.models.encodings import (
+    comp_basis_encoder,
+    entangling_layer,
+    phase_encoder,
+    unary_encoder,
+    unary_encoder_random_gaussian,
 )
+from qibo.models.error_mitigation import CDR, ICS, ZNE, vnCDR
 from qibo.models.evolution import AdiabaticEvolution, StateEvolution
 from qibo.models.grover import Grover
 from qibo.models.qft import QFT
-from qibo.models.qgan import StyleQGAN
 from qibo.models.variational import AAVQE, FALQON, QAOA, VQE
