@@ -1,4 +1,5 @@
 """Models for time evolution of state vectors."""
+
 from qibo import optimizers, solvers
 from qibo.callbacks import Gap, Norm
 from qibo.config import log, raise_error
@@ -54,9 +55,7 @@ class StateEvolution:
         else:
             ham = hamiltonian(0)
             if not isinstance(ham, AbstractHamiltonian):
-                raise TypeError(
-                    "Hamiltonian type {} not understood." "".format(type(ham))
-                )
+                raise TypeError(f"Hamiltonian type {type(ham)} not understood.")
         self.nqubits = ham.nqubits
         self.backend = ham.backend
         if dt <= 0:
@@ -69,8 +68,8 @@ class StateEvolution:
                 raise_error(
                     NotImplementedError,
                     "Distributed evolution is only "
-                    "implemented using the Trotter "
-                    "exponential solver.",
+                    + "implemented using the Trotter "
+                    + "exponential solver.",
                 )
             ham.circuit(dt, accelerators)
         self.solver = solvers.get_solver(solver, self.dt, hamiltonian)
@@ -89,7 +88,10 @@ class StateEvolution:
     def _create_calculate_callbacks(self, accelerators):
         def calculate_callbacks(state):
             for callback in self.callbacks:
-                callback.append(callback.apply(self.backend, state))
+                callback.nqubits = self.nqubits
+                # by executing callbacks.apply we also append the object to history
+                # see callbacks module for this
+                callback.apply(self.backend, state)
 
         if accelerators is None:
             return calculate_callbacks
@@ -236,7 +238,9 @@ class AdiabaticEvolution(StateEvolution):
             )
         self.hamiltonian.total_time = final_time - start_time
         if initial_state is None:
-            initial_state = self.hamiltonian.ground_state()
+            initial_state = self.backend.cast(
+                self.hamiltonian.ground_state(), copy=True
+            )
         return super().execute(final_time, start_time, initial_state)
 
     @staticmethod
