@@ -40,15 +40,18 @@ class Grover:
         .. testcode::
 
             import numpy as np
-            from qibo import gates
-            from qibo.models import Circuit
+
+            from qibo import Circuit, gates
             from qibo.models.grover import Grover
+
             # Create an oracle. Ex: Oracle that detects state |11111>
             oracle = Circuit(5 + 1)
             oracle.add(gates.X(5).controlled_by(*range(5)))
+
             # Create superoposition circuit. Ex: Full superposition over 5 qubits.
             superposition = Circuit(5)
             superposition.add([gates.H(i) for i in range(5)])
+
             # Generate and execute Grover class
             grover = Grover(oracle, superposition_circuit=superposition, number_solutions=1)
             solution, iterations = grover()
@@ -67,7 +70,6 @@ class Grover:
         check_args=(),
         iterative=False,
     ):
-
         self.oracle = oracle
         self.initial_state_circuit = initial_state_circuit
 
@@ -186,10 +188,9 @@ class Grover:
             measured (str): bitstring measured and checked as a valid solution.
             total_iterations (int): number of times the oracle has been called.
         """
-        if backend is None:  # pragma: no cover
-            from qibo.backends import GlobalBackend
+        from qibo.backends import _check_backend
 
-            backend = GlobalBackend()
+        backend = _check_backend(backend)
 
         k = 1
         lamda = lamda_value
@@ -223,10 +224,9 @@ class Grover:
             solution (str): bitstring (or list of bitstrings) measured as solution of the search.
             iterations (int): number of oracle calls done to reach a solution.
         """
-        if backend is None:  # pragma: no cover
-            from qibo.backends import GlobalBackend
+        from qibo.backends import _check_backend
 
-            backend = GlobalBackend()
+        backend = _check_backend(backend)
 
         if (self.num_sol or self.targ_a) and not self.iterative:
             if self.targ_a:
