@@ -72,7 +72,7 @@ class TSP:
             from collections import defaultdict
             from qibo import gates
             from qibo.models import QAOA
-            from qibo.states import CircuitResult
+            from qibo.result import CircuitResult
 
 
             def convert_to_standard_Cauchy(config):
@@ -113,7 +113,7 @@ class TSP:
                 quantum_state = qaoa.execute(initial_state)
                 circuit = Circuit(9)
                 circuit.add(gates.M(*range(9)))
-                result = CircuitResult(small_tsp.backend, circuit, quantum_state, nshots=1000)
+                result = CircuitResult(quantum_state, circuit.measurements, small_tsp.backend, nshots=1000)
                 freq_counter = result.frequencies()
                 # let's combine freq_counter here, first convert each key and sum up the frequency
                 cauchy_dict = defaultdict(int)
@@ -135,12 +135,10 @@ class TSP:
     """
 
     def __init__(self, distance_matrix, backend=None):
-        if backend is None:  # pragma: no cover
-            from qibo.backends import GlobalBackend
+        from qibo.backends import _check_backend
+
+        self.backend = _check_backend(backend)
 
-            self.backend = GlobalBackend()
-        else:
-            self.backend = backend
         self.distance_matrix = distance_matrix
         self.num_cities = distance_matrix.shape[0]
         self.two_to_one = calculate_two_to_one(self.num_cities)
