@@ -1,5 +1,8 @@
+import numpy as np
+
 from qibo.config import raise_error
 from qibo.models.evolution import StateEvolution
+from qibo.models.utils import vqe_loss
 
 
 class VQE:
@@ -33,11 +36,13 @@ class VQE:
         """Initialize circuit ansatz and hamiltonian."""
         self.circuit = circuit
         self.hamiltonian = hamiltonian
+        self.backend = hamiltonian.backend
 
     def minimize(
         self,
         initial_state,
         method="Powell",
+        loss_func=None,
         jac=None,
         hess=None,
         hessp=None,
@@ -57,6 +62,7 @@ class VQE:
             method (str): the desired minimization method.
                 See :meth:`qibo.optimizers.optimize` for available optimization
                 methods.
+            loss (callable): loss function, the default one is :func:`qibo.models.utils.vqe_loss`.
             jac (dict): Method for computing the gradient vector for scipy optimizers.
             hess (dict): Method for computing the hessian matrix for scipy optimizers.
             hessp (callable): Hessian of objective function times an arbitrary
@@ -76,26 +82,20 @@ class VQE:
             the ``OptimizeResult``, for ``'cma'`` the ``CMAEvolutionStrategy.result``,
             and for ``'sgd'`` the options used during the optimization.
         """
-
-        def _loss(params, circuit, hamiltonian):
-            circuit.set_parameters(params)
-            result = hamiltonian.backend.execute_circuit(circuit)
-            final_state = result.state()
-            return hamiltonian.expectation(final_state)
-
+        if loss_func is None:
+            loss_func = vqe_loss
         if compile:
-            loss = self.hamiltonian.backend.compile(_loss)
+            loss = self.hamiltonian.backend.compile(loss_func)
         else:
-            loss = _loss
+            loss = loss_func
 
         if method == "cma":
             # TODO: check if we can use this shortcut
             # dtype = getattr(self.hamiltonian.backend.np, self.hamiltonian.backend._dtypes.get('DTYPE'))
             dtype = self.hamiltonian.backend.np.float64
-            loss = lambda p, c, h: dtype(_loss(p, c, h))
+            loss = lambda p, c, h: dtype(loss_func(p, c, h))
         elif method != "sgd":
-            loss = lambda p, c, h: self.hamiltonian.backend.to_numpy(_loss(p, c, h))
-
+            loss = lambda p, c, h: self.hamiltonian.backend.to_numpy(loss_func(p, c, h))
         result, parameters, extra = self.optimizers.optimize(
             loss,
             initial_state,
@@ -116,6 +116,20 @@ class VQE:
         self.circuit.set_parameters(parameters)
         return result, parameters, extra
 
+    def energy_fluctuation(self, state):
+        """
+        Evaluate energy fluctuation
+
+        .. math::
+            \\Xi_{k}(\\mu) = \\sqrt{\\langle\\mu|\\hat{H}^2|\\mu\\rangle - \\langle\\mu|\\hat{H}|\\mu\\rangle^2} \\,
+
+        for a given state :math:`|\\mu\\rangle`.
+
+        Args:
+            state (np.ndarray): quantum state to be used to compute the energy fluctuation with H.
+        """
+        return self.hamiltonian.energy_fluctuation(state)
+
 
 class AAVQE:
     """This class implements the Adiabatically Assisted Variational Quantum Eigensolver
@@ -167,23 +181,21 @@ class AAVQE:
         bounds_tolerance=1e-7,
         time_tolerance=1e-7,
     ):
-
         if nsteps <= 0:  # pragma: no cover
             raise_error(
                 ValueError,
-                "Number of steps nsteps should be positive but is {}."
-                "".format(nsteps),
+                f"Number of steps nsteps should be positive but is {nsteps}.",
             )
         if t_max <= 0:  # pragma: no cover
             raise_error(
                 ValueError,
-                "Maximum time t_max should be positive but is {}." "".format(t_max),
+                f"Maximum time t_max should be positive but is {t_max}.",
             )
         if easy_hamiltonian.nqubits != problem_hamiltonian.nqubits:  # pragma: no cover
             raise_error(
                 ValueError,
-                "The easy Hamiltonian has {} qubits while problem Hamiltonian has {}."
-                "".format(easy_hamiltonian.nqubits, problem_hamiltonian.nqubits),
+                f"The easy Hamiltonian has {easy_hamiltonian.nqubits} qubits "
+                + f"while problem Hamiltonian has {problem_hamiltonian.nqubits}.",
             )
 
         self.ATOL = bounds_tolerance
@@ -202,7 +214,7 @@ class AAVQE:
             raise_error(
                 ValueError,
                 "Scheduling function must take only one argument,"
-                "but the function proposed takes {}.".format(nparams),
+                + f"but the function proposed takes {nparams}.",
             )
         self.set_schedule(s)
 
@@ -211,10 +223,10 @@ class AAVQE:
         # check boundary conditions
         s0 = func(0)
         if abs(s0) > self.ATOL:  # pragma: no cover
-            raise_error(ValueError, "s(0) should be 0 but it is {}.".format(s0))
+            raise_error(ValueError, f"s(0) should be 0 but it is {s0}.")
         s1 = func(1)
         if abs(s1 - 1) > self.ATOL:  # pragma: no cover
-            raise_error(ValueError, "s(1) should be 1 but it is {}.".format(s1))
+            raise_error(ValueError, f"s(1) should be 1 but it is {s1}.")
         self._schedule = func
 
     def schedule(self, t):
@@ -224,14 +236,12 @@ class AAVQE:
         if (t - self._t_max) > self.ATOL_TIME:  # pragma: no cover
             raise_error(
                 ValueError,
-                "t cannot be greater than {}, but it is {}.".format(self._t_max, t),
+                f"t cannot be greater than {self._t_max}, but it is {t}.",
             )
 
         s = self._schedule(t / self._t_max)
         if (abs(s) - 1) > self.ATOL:  # pragma: no cover
-            raise_error(
-                ValueError, "s cannot be greater than 1 but it is {}.".format(s)
-            )
+            raise_error(ValueError, f"s cannot be greater than 1 but it is {s}.")
         return s
 
     def hamiltonian(self, t):
@@ -239,7 +249,7 @@ class AAVQE:
         if (t - self._t_max) > self.ATOL:  # pragma: no cover
             raise_error(
                 ValueError,
-                "t cannot be greater than {}, but it is {}.".format(self._t_max, t),
+                f"t cannot be greater than {self._t_max}, but it is {t}.",
             )
         # boundary conditions  s(0)=0, s(total_time)=1
         st = self.schedule(t)
@@ -343,9 +353,7 @@ class QAOA:
         self.params = None
         # problem hamiltonian
         if not isinstance(hamiltonian, AbstractHamiltonian):
-            raise_error(
-                TypeError, "Invalid Hamiltonian type {}." "".format(type(hamiltonian))
-            )
+            raise_error(TypeError, f"Invalid Hamiltonian type {type(hamiltonian)}.")
         self.hamiltonian = hamiltonian
         self.nqubits = hamiltonian.nqubits
         # mixer hamiltonian (default = -sum(sigma_x))
@@ -360,16 +368,14 @@ class QAOA:
             if type(mixer) != type(hamiltonian):
                 raise_error(
                     TypeError,
-                    "Given Hamiltonian is of type {} "
-                    "while mixer is of type {}."
-                    "".format(type(hamiltonian), type(mixer)),
+                    f"Given Hamiltonian is of type {type(hamiltonian)} "
+                    + f"while mixer is of type {type(mixer)}.",
                 )
             if mixer.nqubits != hamiltonian.nqubits:
                 raise_error(
                     ValueError,
-                    "Given Hamiltonian acts on {} qubits "
-                    "while mixer acts on {}."
-                    "".format(hamiltonian.nqubits, mixer.nqubits),
+                    f"Given Hamiltonian acts on {hamiltonian.nqubits} qubits "
+                    + f"while mixer acts on {mixer.nqubits}.",
                 )
             self.mixer = mixer
 
@@ -381,8 +387,8 @@ class QAOA:
             raise_error(
                 NotImplementedError,
                 "Distributed QAOA is implemented "
-                "only with SymbolicHamiltonian and "
-                "exponential solver.",
+                + "only with SymbolicHamiltonian and "
+                + "exponential solver.",
             )
         if isinstance(self.hamiltonian, self.hamiltonians.SymbolicHamiltonian):
             self.hamiltonian.circuit(1e-2, accelerators)
@@ -452,7 +458,8 @@ class QAOA:
         initial_p,
         initial_state=None,
         method="Powell",
-        mode=None,
+        loss_func=None,
+        loss_func_param=dict(),
         jac=None,
         hess=None,
         hessp=None,
@@ -476,8 +483,8 @@ class QAOA:
             method (str): the desired minimization method.
                 See :meth:`qibo.optimizers.optimize` for available optimization
                 methods.
-            mode (str): the desired loss function. The default is None. Alternatives are
-             "cvar", and "gibbs".
+            loss_func (function): the desired loss function. If it is None, the expectation is used.
+            loss_func_param (dict): a dictionary to pass in the loss function parameters.
             jac (dict): Method for computing the gradient vector for scipy optimizers.
             hess (dict): Method for computing the hessian matrix for scipy optimizers.
             hessp (callable): Hessian of objective function times an arbitrary
@@ -502,21 +509,22 @@ class QAOA:
             .. testcode::
 
                 from qibo import hamiltonians
+                from qibo.models.utils import cvar, gibbs
 
                 h = hamiltonians.XXZ(3)
                 qaoa = models.QAOA(h)
                 initial_p = [0.314, 0.22, 0.05, 0.59]
                 best, params, _ = qaoa.minimize(initial_p)
-                best, params, _ = qaoa.minimize(initial_p, mode="gibbs")
-                best, params, _ = qaoa.minimize(initial_p, mode="cvar")
+                best, params, _ = qaoa.minimize(initial_p, loss_func=cvar, loss_func_param={'alpha':0.1})
+                best, params, _ = qaoa.minimize(initial_p, loss_func=gibbs, loss_func_param={'eta':0.1})
 
         """
         if len(initial_p) % 2 != 0:
             raise_error(
                 ValueError,
                 "Initial guess for the parameters must "
-                "contain an even number of values but "
-                "contains {}.".format(len(initial_p)),
+                + "contain an even number of values but "
+                + "contains {len(initial_p)}.",
             )
 
         def _loss(params, qaoa, hamiltonian, state):
@@ -524,16 +532,17 @@ class QAOA:
                 state = hamiltonian.backend.cast(state, copy=True)
             qaoa.set_parameters(params)
             state = qaoa(state)
-            if mode is None:
+            if loss_func is None:
                 return hamiltonian.expectation(state)
-            elif mode == "cvar":
-                from qibo.models.utils import cvar
+            else:
+                func_hyperparams = {
+                    key: loss_func_param[key]
+                    for key in loss_func_param
+                    if key in loss_func.__code__.co_varnames
+                }
+                param = {**func_hyperparams, "hamiltonian": hamiltonian, "state": state}
 
-                return cvar(hamiltonian, state)
-            elif mode == "gibbs":
-                from qibo.models.utils import gibbs
-
-                return gibbs(hamiltonian, state)
+                return loss_func(**param)
 
         if method == "sgd":
             loss = lambda p, c, h, s: _loss(self.hamiltonian.backend.cast(p), c, h, s)
@@ -623,8 +632,6 @@ class FALQON(QAOA):
             The corresponding best parameters.
             extra: variable with historical data for the energy and callbacks.
         """
-        import numpy as np
-
         parameters = np.array([delta_t, 0])
 
         def _loss(params, falqon, hamiltonian):
@@ -634,7 +641,7 @@ class FALQON(QAOA):
 
         energy = [np.inf]
         callback_result = []
-        for it in range(1, max_layers + 1):
+        for _ in range(1, max_layers + 1):
             beta = self.hamiltonian.backend.to_numpy(
                 _loss(parameters, self, self.evol_hamiltonian)
             )
